module LifePex::AchivementDSL
  AVAILABLE_GENERIC_SUCCESS = {}
  AVAILABLE_SETUP_SUCCESS = {}
  UNDEFINED = :undefined_by_dsl

  def self.dslize_input(name)
    define_method name do |v = UNDEFINED|
      if v == :undefined_by_dsl
        instance_variable_get "@#{name}"
      else
        instance_variable_set "@#{name}", v
      end
    end
  end

  def self.dslize_flag(name)
    define_method "is_#{name}" do
      instance_variable_get "@#{name}"
    end

    define_method name do
      instance_variable_set "@#{name}", true
    end
  end

  class Success
    class Parameter
      def initialize(name, type = nil, &block)
        @name = name
        @type = type
        @description = nil
        instance_eval(&block) if block_given?
      end

      LifePex::AchivementDSL.dslize_input :name
      LifePex::AchivementDSL.dslize_input :type
      LifePex::AchivementDSL.dslize_input :description
      LifePex::AchivementDSL.dslize_flag :required

      # not dsl related
      def accept?(v)
        return true if @input_type.nil?
        return true if @input_type == :Numeric && v.is_a?(Numeric)
        return true if @input_type == :Integer && v.is_a?(Numeric) && v.round(0) == v
        return true if @input_type == :String && v.is_a?(String)
        return true if @input_type == :Array && v.is_a?(Array)
        return false
      end
    end

    class Setup
      def initialize(success, name, &block)
        @success = success
        @name = name
        @parameters = {}
        @icon = nil
        instance_eval(&block)
      end

      def parameter(**kv)
        param_name = kv.keys.first
        param_value = kv.values.first
        param = @success.parameters.find { |param| param.name == param_name }
        raise RuntimeError.new "Invalid value #{param_value} for #{param_name}" unless param_value.nil? && !param.is_required || param.accept?(param_value)
        @parameters[param_name] = param_value
      end

      def icon(value = UNDEFINED)
        @icon = value if value != UNDEFINED
        if @icon == :gold
          "&#129351;"
        elsif @icon == :silver
          "&#129352;"
        elsif @icon == :copper
          "&#129353;"
        else
          "&#127941;"
        end
      end

      def call(user_pexs, pexs)
        to_complete = @success.get_to_complete
        instance_exec user_pexs, pexs, &to_complete
      end

      def given(name)
        @parameters[name]
      end
    end

    attr_reader :parameters
    LifePex::AchivementDSL.dslize_input :name

    def initialize(name, &block)
      @parameters = []
      @to_complete = nil
      instance_eval(&block) if block_given?
    end

    def parameter(name, type = nil, &block)
      @parameters << Parameter.new(name, type, &block)
    end

    def to_complete(&block)
      @to_complete = block
    end

    def get_to_complete
      @to_complete
    end

    def setup(name, &block)
      if block_given?
        s = Setup.new(self, name, &block)
        LifePex::AchivementDSL::AVAILABLE_SETUP_SUCCESS[name] = s
        s
      else
        LifePex::AchivementDSL::AVAILABLE_SETUP_SUCCESS[name]
      end
    end
  end

  def success(name, &block)
    if block_given?
      s = Success.new(name, &block)
      LifePex::AchivementDSL::AVAILABLE_GENERIC_SUCCESS[name] = s
      s
    else
      LifePex::AchivementDSL::AVAILABLE_GENERIC_SUCCESS[name]
    end
  end
end
